package sis.util;

import junit.framework.TestCase;
import java.util.concurrent.locks.*;
import java.util.concurrent.TimeUnit;

public class ThreadPoolTest extends TestCase {
   private Lock lock;
   private ThreadPool pool;
   private Condition start;

   protected void setUp() {
      lock = new ReentrantLock();
      final int poolSize = 2;
      pool = new ThreadPool(poolSize);
      start = lock.newCondition();
   }

   public void testSingleThread() throws InterruptedException {
      Condition workCompleted = lock.newCondition();
      pool.add(createRunnable(workCompleted));
      System.out.println("-1-");
      signalAll(start);
      System.out.println("-2-");

      lock.lock();
      try {
         assertTrue(workCompleted.await(5000, TimeUnit.MILLISECONDS));
      }
      finally {
         lock.unlock();
      }

   }

   public void xtestMultipleThreadsLessThanPoolSize() throws InterruptedException {
      Condition thread1Completed = lock.newCondition();
      Condition thread2Completed = lock.newCondition();

      pool.add(createRunnable(thread1Completed));
      pool.add(createRunnable(thread2Completed));

      verify(thread1Completed);
      verify(thread2Completed);
   }

   public void xtestMultipleThreadsGreaterThanPoolSize() throws InterruptedException {
      final Condition thread1Completed = lock.newCondition();
      final Condition thread2Completed = lock.newCondition();
      final Condition thread3Completed = lock.newCondition();

      final Condition startInitialThreads = lock.newCondition();
      final Condition startExtraThread = lock.newCondition();

      Runnable runnable1 = new Runnable() {
         public void run() {
            await(startInitialThreads);
            signalAll(thread1Completed);
         }
      };

      Runnable runnable2 = new Runnable() {
         public void run() {
            await(startInitialThreads);
            signalAll(thread2Completed);
         }
      };

      Runnable runnable3 = new Runnable() {
         public void run() {
            await(startExtraThread);
            signalAll(thread3Completed);
         }
      };

      pool.add(runnable1);
      pool.add(runnable2);
      pool.add(runnable3);

      setToVerify(thread1Completed);
      setToVerify(thread2Completed);
      setToVerify(thread3Completed);

      lock.lock();
        try {
           ReentrantLock lockImpl = (ReentrantLock)lock;
           assertTrue(lockImpl.hasWaiters(startInitialThreads));
           assertFalse(lockImpl.hasWaiters(startExtraThread));
        }
        finally {
           lock.unlock();
        }

      signalAll(startInitialThreads);
      signalAll(startExtraThread);

      // test completes before verification threads complete!
System.out.println("--");
   }

   private void sleep(long ms) {
      try {
         Thread.sleep(ms);
      } catch (InterruptedException e) {
      }
   }

   private void setToVerify(final Condition condition) {
      new Thread() {
         public void run() {
            try {
               verify(condition);
            } catch (InterruptedException e) {
            }
         }
      }.start();
   }

   private void verify(Condition condition) throws InterruptedException {
      lock.lock();
      try {
         assertTrue(condition.await(5000, TimeUnit.MILLISECONDS));
      }
      finally {
         lock.unlock();
      }
   }

   private Runnable createRunnable(final Condition threadCompleted) {
      return new Runnable() {
         public void run() {
System.out.println("-a-");
            await(start);
System.out.println("-b-");
            signalAll(threadCompleted);
System.out.println("-c-");
         }
      };
   }

   private void await(Condition condition) {
      lock.lock();
      try {
         condition.await();
      }
      catch (InterruptedException e) {
         fail("unexpected interruption");
      }
      finally {
         lock.unlock();
      }
   }

   private void signalAll(Condition condition) {
      lock.lock();
      try {
         condition.signalAll();
      }
      finally {
         lock.unlock();
      }
   }
}
